#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <stddef.h>
#include <cerrno>
#include <unistd.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <pthread.h>
#include <string>

#define MSG_LEN 200

#define SOCKET_ADDR "/var/run/adslserv.sock"

using std::string;

int modem_status = 0;
int modem_type = 0;
int clientsock;
const int max_devices = 5;
string devices[max_devices+1];

typedef struct {
	char 	* user;
	char 	* pass;
	char 	* type;
	int	vpi;
	int	vci;
	char	* path;
} Modem;

void adsl_connect( )
{

	int child;

	switch( modem_type ) 
	{
		case 1:
			child = fork( );
	
			if( child == 0 )
				execl( "/usr/sbin/pppd", "pppd", "call", "gprs", 0 );
			break;

		case 2:
		case 3:	
			child = fork( );

			if( child == 0 )
				execl( "/usr/sbin/adsl-start", "adsl-start", 0 );

			break;
	}
}


int adsl_disconnect( )
{

	int e;

	switch( modem_type )
	{
		case 1:
			system( "killall pppd" );
			wait3( 0, 0, 0 );
			return 1;
			break;

		case 2:
		case 3:
			e = system( "/usr/sbin/adsl-stop" );
			wait3( 0, 0, 0 );
			if( WEXITSTATUS( e ) != 0 )
				return 0;
			return 1;	

	}

	return 1;
}

int adsl_status( )
{
	int e;
	e = system( "pidof pppd >/dev/null" );
	if( WEXITSTATUS( e ) != 0 )
		return 0;
	else return 1;
}

// return values:
// 0 - connecting...
// 1 - connected
// 2 - not connected

int adsl_new_connection( )
{
	int e, ex;
	e = system( "ifconfig | grep ppp0 >/dev/null" );
	
	if( WEXITSTATUS( e ) == 0  )
		return 1;

	else
	{
		ex = wait3( 0, WNOHANG, 0 );
		if( ex != 0 )
			return 2;
		else return 0;
	}

}

void non_exit_error( char * err )
{
//	char * buff;
//	buff = strerror( errno );
//	fprintf( stderr, err );
//	fprintf( stderr, " %s\n", buff );
}

void error( char * err )
{
//	non_exit_error( err );
//	exit( 1 );
}

int check_modem( )
{
	struct stat st;

	if( stat( "/proc/net/atm/speedtch:0", &st ) == 0 )
	{
		modem_type = 3;
		modem_status = 1;
	       	return 1;
	}
	if( stat( "/proc/net/atm/ueagle-atm:0", &st ) == 0 )
	{
		modem_type = 2;
		modem_status = 1;
		return 1;
	}

	if( stat( "/dev/noz0", &st ) == 0 )
	{
		modem_type = 1;
		modem_status = 1;
		return 1;
	}

	return 0;

//	return modem_status;
}

int check_window( )
{
	struct stat st;

	switch( modem_type )
	{
		case 1:
			if( stat( "/etc/ppp/peers/gprs", &st ) == -1 )
				return modem_type;
			break;
			
		case 2:
		case 3:
			if( stat( "/etc/sysconfig/network-scripts/ifcfg-ppp0", &st ) == -1 )
				return modem_type;
			break;
	}

	return 0;
}

int check_modem_type( )
{
	return modem_type;
}

void send_int( int data )
{
	char buff[MSG_LEN];
	int e;

	memset( buff, 0, MSG_LEN );
	snprintf( buff, MSG_LEN, "%d", data );
	e = send( clientsock, buff, strlen( buff ), MSG_NOSIGNAL );

	if( e < 0 )
		non_exit_error( "Could not send data" );
}

void send_command( char * data )
{
	int e;

	e = send( clientsock, data, strlen( data ), MSG_NOSIGNAL );

	if( e < 0 )
		non_exit_error( "Could not send data" );
}

static DBusHandlerResult signal_filter( DBusConnection * conn, DBusMessage * msg, void * user_data )
{
	GMainLoop * loop_new = (GMainLoop *)user_data;

	if( dbus_message_is_signal( msg, "org.freedesktop.Hal.Manager", "DeviceAdded" ) )
	{
		DBusError err;
		char * s;
		int i = 1;
		int a;

		dbus_error_init( &err );

		if( dbus_message_get_args( msg, &err, DBUS_TYPE_STRING, &s, DBUS_TYPE_INVALID ) )
		{

			for( ; i <= max_devices; i++ )
			{
				if( strncmp( devices[i].c_str( ), s, strlen( devices[i].c_str( ) ) ) == 0 &&
					( rindex( s, '_' ) - s ) == devices[i].length( ) )
				{
					modem_status = 1;
					modem_type = i;

					if( i == 4 || i == 5 )
						modem_type = 2;
					
					a = check_window( );
					if( a == 1 ) send_command( "WINDOW 1" );
					if( a == 2 ) send_command( "WINDOW 2" );
					if( a == 3 ) send_command( "WINDOW 3" );

					if( a == 0 )
						send_command( "MODEM 1" );
				}

			}	

		}
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	if( dbus_message_is_signal( msg, "org.freedesktop.Hal.Manager", "DeviceRemoved" ) )
        {
                DBusError err;
                char * s;
		int i = 1;
                dbus_error_init( &err );

                if( dbus_message_get_args( msg, &err, DBUS_TYPE_STRING, &s, DBUS_TYPE_INVALID ) )
                {
			for( ; i <= max_devices; i++ )
			{
				if( strncmp( devices[i].c_str( ), s, strlen( devices[i].c_str( ) ) ) == 0 &&
					( rindex( s, '_' ) - s ) == devices[i].length( ) )
				{
					adsl_disconnect( );
					modem_status = 0;
					modem_type = -1;
					send_command( "MODEM 0" );
				}
			}
                }
                return DBUS_HANDLER_RESULT_HANDLED;
        }

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

}

void * run_dbus( void * arg )
{
	GMainLoop * loop;
	DBusConnection * bus;
	DBusError err;

	devices[1] = "/org/freedesktop/Hal/devices/pci_1931_c";
	devices[3] = "/org/freedesktop/Hal/devices/usb_device_6b9_4061";
	devices[2] = "/org/freedesktop/Hal/devices/usb_device_1110_9021";
	devices[4] = "/org/freedesktop/Hal/devices/usb_device_1110_9031";
	devices[5] = "/org/freedesktop/Hal/devices/usb_device_1110_900f";
	

	loop = g_main_loop_new( NULL, FALSE );
	dbus_error_init( &err );
	bus = dbus_bus_get( DBUS_BUS_SYSTEM, &err );	

	if( ! bus )
	{
//		error( err.message );
//		dbus_error_free( &err );
	}

	dbus_connection_setup_with_g_main( bus, NULL );
	
	dbus_bus_add_match( bus, "type='signal',interface='org.freedesktop.Hal.Manager'", &err );
	dbus_connection_add_filter( bus, signal_filter, loop, NULL );
	g_main_loop_run( loop );
	return 0;
}

Modem * modem_init( char * msg )
{
	char * type;
	char * user;
	char * pass;
	char * path;
	char * space;
	char * vpi;
	char * vci;

	Modem * m;	

	msg += 5;

	type = msg;
	space = index( msg, ' ' );
	*space = 0;
	user = ++space;
	space = index( space, ' ' );
	*space = 0;
	pass = ++space;
	space = index( space, ' ' );
	*space = 0;
	vpi = ++space;
	space = index( space, ' ' );
	*space = 0;
	vci = ++space;
	space = index( space, ' ' );
	*space = 0;
	path = ++space;

	
	m = (Modem*)malloc( sizeof( Modem ) );
	m->type = type;
	m->user = user;
	m->pass = pass;
	m->vpi = atoi( vpi );
	m->vci = atoi( vci );
	m->path = path;

	return m;
}

int main( )
{
	int s, result, e;
	struct sockaddr_un name;
	socklen_t name_len;
	struct stat st;
	char msg[MSG_LEN];
	char text1[] = "CONNECT";
	char text2[] = "DISCONNECT";
	char text3[] = "STATUS";
	char text4[] = "MODEM";
	char text5[] = "DISCONNECT_QUIT";
	char text6[] = "CONNECTION";
	char text9[] = "WINDOW";
	char text10[] = "GPRS";
	char text11[] = "MODEM_TYPE";
	char text12[] = "ADSL";
	char text13[] = "SHARE_DHCP";
	char text14[] = "SHARE_NODHCP";
	char text15[] = "SHARE_QUIT_DHCP";
	char text16[] = "SHARE_QUIT_NODHCP";
	int modem_type = 0;

	pthread_t upTh = 0;

	char bufor[256];
        pid_t PID=getpid();
        sprintf(bufor,"/sbin/pidof 'adsl-manager-daemon' -o %d 1> /dev/null 2> /dev/null < /dev/null",PID);
        int ex = system(bufor);
        if( WEXITSTATUS( ex )==0)
        {
                return 6;
        }

	if( fork( ) ) exit( 0 );

	pthread_create( &upTh, 0, run_dbus, 0 );

	if( stat( SOCKET_ADDR, &st ) == 0 )
		unlink( SOCKET_ADDR );
	
	
	s = socket( PF_LOCAL, SOCK_STREAM, 0 );

	if( s < 0 )
	{
//		fprintf( stderr, "Could not create socket\n" );
		return 1;
	}

	name.sun_family = AF_LOCAL;
	strcpy( name.sun_path, SOCKET_ADDR );

	if( bind( s, ( struct sockaddr *)&name, sizeof( name ) ) < 0 )
		error( "Could not bind" );

	chmod( SOCKET_ADDR, S_IRUSR | S_IWUSR | S_IROTH | S_IWOTH );

	listen( s, 5 );

	modem_status = check_modem( );
	
	name_len = SUN_LEN( &name );

	while( clientsock = accept( s, ( struct sockaddr * )&name, &name_len ) )
	{

		int last_errno = errno;
		
		modem_status = check_modem( );
		
		if( clientsock < 0 )
		{
			if( last_errno == 11 )
				continue;
			else
				error( "Error while accepting" );
		}
		
		memset( msg, 0, MSG_LEN );
		
		while( ( result = read( clientsock, msg, MSG_LEN ) ) )
		{
			if( result <= 0 )
			{
				if( errno == 11 )
				{
					modem_status = check_modem( );
					continue;
				}
				else 
				{
					non_exit_error( "Error while receiving data" );
					break;
				}
			}
			

	//		printf( "received: %s\n", msg );

			if( strcmp( msg, text1 ) == 0 )
			{
				adsl_connect( );
				send_int( 1 );
			}

			else if( strcmp( msg, text2 ) == 0 )
			{
				e = adsl_disconnect( );
				send_int( e );
			}

			else if( strcmp( msg, text3 ) == 0 )
			{
				e = adsl_status( );
				send_int( e );
			}

			else if( strcmp( msg, text4 ) == 0 )
			{
				int a;
				e = check_modem( );
				send_int( e );
				a = check_window( );
				sleep( 1 );
				if( a == 1 ) send_command( "WINDOW 1" );
				if( a == 2 ) send_command( "WINDOW 2" );
				if( a == 3 ) send_command( "WINDOW 3" );
			}

			else if( strcmp( msg, text5 ) == 0 )
			{
				adsl_disconnect( );
			}

			else if( strcmp( msg, text6 ) == 0 )
			{
				e = adsl_new_connection( );
				send_int( e );
			}

			else if( strcmp( msg, text9 ) == 0 )
			{
				e = check_window( );
				send_int( e );
			}

                        else if( strcmp( msg, text11 ) == 0 )
                        {
                                e = check_modem_type( );
                                send_int( e );
                        }

	
			else if( strncmp( msg, text10, 4 ) == 0 )
			{
				char * space1;
				char * space2;	
				char cmd[50];
				int pin;

				space1 = strchr( msg, ' ' );
				space1++;
				space2 = strchr( space1, ' ' );

				*space2 = 0;
				pin = atoi( space1 );
				
				snprintf( cmd, 50, "%s %d %s", "/usr/share/nozomi/install_gprs.sh", pin, space2 );

				system( cmd );

				modem_type = 1;
				
				send_int( 1 );			
	
			}

			else if( strncmp( msg, text12, 4 ) == 0 )
			{
			
				char cmd[1024];
				Modem * modem = modem_init( msg );

				snprintf( cmd, 1024, "%s %s %s %s %d %d %s", "/usr/share/adsl-manager-daemon/adsl_install", modem->type, modem->user, modem->pass, modem->vci, modem->vpi, modem->path );
				system( cmd );
				
				if( strcmp( modem->type, "sagem" ) == 0 )
					modem_type = 2;
				if( strcmp( modem->type, "speedtouch" ) == 0 )
					modem_type = 3;

				send_int( 1 );
			}
			
			else if( strcmp( msg, text13 ) == 0 )
			{
				e = system( "/usr/share/adsl-manager-daemon/start_dhcpd.sh" );
				ex = system( "iptables -t nat -A POSTROUTING -o ppp0 -j MASQUERADE" );
				int z = system( "echo 1 > /proc/sys/net/ipv4/ip_forward" );

				if( WEXITSTATUS( e ) != 0 && WEXITSTATUS( ex ) != 0 && WEXITSTATUS( z ) != 0 )
					send_int( 0 );
				else
					send_int( 1 );
						
			}

			else if( strcmp( msg, text14 ) == 0 )
			{	
				e = system( "iptables -t nat -A POSTROUTING -o ppp0 -j MASQUERADE" );
				ex = system( "echo 1 > /proc/sys/net/ipv4/ip_forward" );
	
				if( WEXITSTATUS( e ) == 0 && WEXITSTATUS( ex ) == 0 )
					send_int( 1 );
				else
					send_int( 0 );
				

			}
	
			else if( strcmp( msg, text15 ) == 0 )
			{
				e = system( "/usr/share/adsl-manager-daemon/stop_dhcpd.sh" );
				ex = system ( "iptables -t nat -D POSTROUTING -o ppp0 -j MASQUERADE" );
				int z = system( "echo 0 > /proc/sys/net/ipv4/ip_forward" );

				if( WEXITSTATUS( e ) != 0 && WEXITSTATUS( ex ) != 0 && WEXITSTATUS( z ) != 0 )
					send_int( 0 );
				else
					send_int( 1 );
			}

			else if( strcmp( msg, text16 ) == 0 )
			{
				e = system( "iptables -t nat -D POSTROUTING -o ppp0 -j MASQUERADE" );
				ex = system( "echo 0 > /proc/sys/net/ipv4/ip_forward" );
				
				if( WEXITSTATUS( e ) == 0 && WEXITSTATUS( ex ) == 0 )
					send_int( 1 );
				else
					send_int( 0 );
			}

			memset( msg, 0, MSG_LEN );
		}
	}

	close( s );
	unlink( SOCKET_ADDR );
	
	return 0;

	
	
}



